// Copyright 2019-2024 Yuji Tanaka
// HP : http://galaxystar.image.coocan.jp/
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
/*
 * 3D maze
 * 
 * File  wall.dat
 *       maze?.dat
 * 
 * 2019' 08/15 Yuji Tanaka
 * 2024' 06/02 Yuji Tanaka
 *   For ESP32-S3.
 */
#include "FS.h"
#include "SD.h"
#include <SPI.h>
#include <ILI9341_game.h>
#include "pitches.h"

#define SD_CS 10
#define TFT_DC 48
#define TFT_CS 47
#define TFT_RST 45

ILI9341_game tft = ILI9341_game(TFT_CS, TFT_DC, TFT_RST);

int _width, _height;

int StickXPin = 6;
int StickYPin = 7;
int StickX;
int StickY;

int ButtonAPin = 15;
int ButtonBPin = 16;
int ButtonA;
int ButtonB;

int SoundPin = 17;

#define SCREEN_WIDTH  240
#define SCREEN_HEIGHT 240
#define SCREEN_START_X 0
#define SCREEN_START_Y 64

struct POINT {
  int x, y;
};

struct f_vec {
  double x,y,z;
};

/*  迷路データ */
int wall[22][16];

/*  壁データ  */
struct f_wall{
  int x,y,z,typ;
};
struct f_wall *wall_dat;
struct f_wall *wallpt[256];

/*  polygonデータ  */
struct f_polygon {
  POINT pnt[8];
  int n;
  uint16_t c;
  double dmax;
  double dmin;
}*polygon_dat;
int poly_num;

void show_maze(int,int,int);
void draw_wall(int,int,int,int,int,struct f_vec *,struct f_vec *);
void push_polygon(POINT *, int, uint16_t, double, double);
void draw_polygon();
struct f_vec *pos3d(struct f_vec *,struct f_vec *,struct f_vec *);
struct f_vec *vec_add(struct f_vec *,struct f_vec *,struct f_vec *);
struct f_vec *vec_sub(struct f_vec *,struct f_vec *,struct f_vec *);

int state;
int stage;
int my_x, my_y, my_dir;
int start_x,start_y;
int goal_x,goal_y;
int count;
int draw_count;

void setup()
{
    int x, y;

    Serial.begin(115200);

    Serial.print("Initializing SD card...");
//    pinMode(SD_CS, PULLUP);
//    pinMode(MISO, PULLUP);
//    if (!SD.begin(SD_CS)) {
    if (!SD.begin()) {
      Serial.println("failed!");
    }
    Serial.println("OK!");

    tft.begin(40000000);
    tft.setRotation(0);

// Line inversion
//    tft.startWrite();
//    tft.writeCommand(0xb4);
//    tft.spiWrite(0x00);
//    tft.endWrite();

    tft.clearDisplay();
    _width=tft.width();
    _height=tft.height();
    tft.display();

    pinMode(ButtonAPin, INPUT_PULLUP);
    pinMode(ButtonBPin, INPUT_PULLUP);
  
    state=0;
    stage=0;
    my_dir=0;
    count=0;
    draw_count=0;

    wall_dat=(f_wall *)malloc(sizeof(struct f_wall)*512);
    polygon_dat=(f_polygon *)malloc(sizeof(struct f_polygon)*256);

    load_wall();
}

void loop(){
    int n;

    StickX = analogRead(StickXPin)/4;
    StickY = analogRead(StickYPin)/4;
    ButtonA = digitalRead(ButtonAPin);
    ButtonB = digitalRead(ButtonBPin);

    // Title
    if (state==0){
      show_title(count*5);
      if (ButtonA==0)
        state=1;

    // Initialize
    }else if (state==1){
      load_maze(stage);
      state=2;
      draw_count=0;
      tone(SoundPin, NOTE_C5, 50);
      delay(50);
      tone(SoundPin, NOTE_E5, 50);
      delay(50);
      tone(SoundPin, NOTE_G5, 50);
      delay(50);
      tone(SoundPin, NOTE_C6, 50);
      delay(50);

    // Game
    }else if (state==2){
      // Not move
      if (draw_count==0){
        // Turn left
        if (StickX<256){
          my_dir--;
          if(my_dir<0)
            my_dir=3;
          draw_count=299;
        // Turn right
        }else if (StickX>768){
          my_dir++;
          if(my_dir>3)
            my_dir=0;
          draw_count=399;
        // Go front
        }else if (StickY>768){
          switch(my_dir){
            case 0:
              if(wall[my_y-1][my_x]<10){
                my_y--;
                draw_count=99;
              }
              break;
            case 1:
              if(wall[my_y][my_x+1]<10){
                my_x++;
                draw_count=99;
              }
              break;
            case 2:
              if(wall[my_y+1][my_x]<10){
                my_y++;
                draw_count=99;
              }
              break;
            case 3:
              if(wall[my_y][my_x-1]<10){
                my_x--;
                draw_count=99;
              }
              break;
          }
        // Go back
        }else if (StickY<256){
          switch(my_dir){
            case 0:
              if(wall[my_y+1][my_x]<10){
                my_y++;
                draw_count=199;
              }
              break;
            case 1:
              if(wall[my_y][my_x-1]<10){
                my_x--;
                draw_count=199;
              }
              break;
            case 2:
              if(wall[my_y-1][my_x]<10){
                my_y--;
                draw_count=199;
              }
              break;
            case 3:
              if(wall[my_y][my_x+1]<10){
                my_x++;
                draw_count=199;
              }
              break;
          }
        }
      }else{
        draw_count-=20;
        if(draw_count<0 || draw_count % 100<20)
          draw_count=0;
      }
      show_maze(draw_count);
      // Goal
      if(draw_count==0 && my_x==goal_x && my_y==goal_y){
        state=3;
      }

    // Goal
    }else if (state==3){
      tone(SoundPin, NOTE_G5, 50);
      delay(50);
      tone(SoundPin, NOTE_C6, 50);
      delay(50);
      tone(SoundPin, NOTE_G5, 50);
      delay(50);
      tone(SoundPin, NOTE_C6, 50);
      delay(50);
      state=4;

    // Clear
    }else if (state==4){
      tft.setTextSize(4);
      tft.setTextColor(ILI9341_CYAN);
      tft.setCursor(SCREEN_START_X+SCREEN_WIDTH/2-48, SCREEN_START_Y+SCREEN_HEIGHT/2-8);
      tft.print("Goal");
      tft.display();
      if (ButtonA==0){
        stage++;
        if (stage>6)
          stage=0;
        state=1;
      }
    }

    count++;
//    delay(30);
    noTone(SoundPin);
}

void show_title(int num)
{
  struct f_vec my_pos;
  struct f_vec my_eye;

  tft.clearDisplay();

  my_pos.x=-cos(M_PI*(double)num/200)*10;
  my_pos.y=-sin(M_PI*(double)num/200)*10;
  my_pos.z=-cos(M_PI*(double)num/100)*3;
  my_eye.x=cos(M_PI*(double)num/200);
  my_eye.y=sin(M_PI*(double)num/200);
  my_eye.z=cos(M_PI*(double)num/100)*3/10;

  push_polygon(NULL,0,0,0,0);

  draw_wall(90,0,0,0,0,&my_pos,&my_eye);

  draw_polygon();

  tft.fillRect(0, 0, _width, SCREEN_START_Y, ILI9341_BLACK);
  tft.fillRect(0, 0, SCREEN_START_X, _height, ILI9341_BLACK);
  tft.fillRect(0, SCREEN_START_Y+SCREEN_HEIGHT, _width, _height-(SCREEN_START_Y+SCREEN_HEIGHT), ILI9341_BLACK);
  tft.drawRect(SCREEN_START_X, SCREEN_START_Y, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, ILI9341_WHITE);

  tft.setTextSize(4);
  tft.setTextColor(ILI9341_WHITE);
  tft.setCursor(32, 4);
  tft.print("3D maze");

  tft.display();
}

void show_maze(int num)
{
  int i, j;
  int wall_no;
  struct f_vec wall_xyz;
  struct f_vec my_pos;
  struct f_vec my_eye;
  struct f_vec vis_pos;
  struct f_vec tmp;
  
  tft.clearDisplay();

  tft.fillRect(SCREEN_START_X, SCREEN_START_Y, _width, SCREEN_WIDTH/2-8, 0x39E7);
  tft.fillRect(SCREEN_START_X, SCREEN_START_Y+SCREEN_WIDTH/2+8, _width, SCREEN_WIDTH/2, 0x39E7);

  my_pos.x=0;
  my_pos.y=-1;
  my_pos.z=0;
  my_eye.x=0;
  my_eye.y=1;
  my_eye.z=0;
  if(0<num && num<=100){
    my_pos.x=0;
    my_pos.y=-(double)num/10;
    my_pos.z=0;
    my_eye.x=0;
    my_eye.y=1;
    my_eye.z=0;
  }
  if(100<num && num<=200){
    my_pos.x=0;
    my_pos.y=((double)num-100)/10;
    my_pos.z=0;
    my_eye.x=0;
    my_eye.y=1;
    my_eye.z=0;
  }
  if(200<num && num<=300){
    my_pos.x=0;
    my_pos.y=0;
    my_pos.z=0;
    my_eye.x=cos(M_PI*(300-(double)num)/200);
    my_eye.y=sin(M_PI*(300-(double)num)/200);
    my_eye.z=0;
  }
  if(300<num && num<=400){
    my_pos.x=0;
    my_pos.y=0;
    my_pos.z=0;
    my_eye.x=-cos(M_PI*(400-(double)num)/200);
    my_eye.y=sin(M_PI*(400-(double)num)/200);
    my_eye.z=0;
  }
  push_polygon(NULL,0,0,0,0);

  for(i=-3;i<=3;i++){
    for(j=-3;j<=3;j++){
      if(my_dir==0) wall_no=wall[my_y-j][my_x+i];
      if(my_dir==1) wall_no=wall[my_y+i][my_x+j];
      if(my_dir==2) wall_no=wall[my_y+j][my_x-i];
      if(my_dir==3) wall_no=wall[my_y-i][my_x-j];
      if(wall_no>0){
        draw_wall(wall_no,i*10,j*10,0,my_dir,&my_pos,&my_eye);
      }
    }
  }

  draw_polygon();

  tft.fillRect(0, 0, _width, SCREEN_START_Y, ILI9341_BLACK);
  tft.fillRect(0, 0, SCREEN_START_X, _height, ILI9341_BLACK);
  tft.fillRect(0, SCREEN_START_Y+SCREEN_HEIGHT, _width, _height-(SCREEN_START_Y+SCREEN_HEIGHT), ILI9341_BLACK);
  tft.drawRect(SCREEN_START_X, SCREEN_START_Y, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, ILI9341_WHITE);

  tft.setTextSize(4);
  tft.setTextColor(ILI9341_WHITE);
  tft.setCursor(32, 4);
  tft.print("3D maze");

  tft.display();
}

void draw_wall(int no,int x,int y,int z,int dir,struct f_vec *pos,struct f_vec *eye)
{
  int i;
  struct f_wall *tmppt;
  struct f_vec wall_xyz;
  struct f_vec vis_pos;
  struct f_vec tmp;
  int back;
  double dirmax,dirmin;
  POINT aPoints[8];

  tmppt=wallpt[no];
  while(tmppt->typ>=0){
    back=0;
    dirmax=0;
    dirmin=10000000;
    for(i=0;tmppt->typ==0;i++){
    if(dir==0){
      wall_xyz.x=(double)(tmppt->x+x);
      wall_xyz.y=(double)(tmppt->y+y);
      wall_xyz.z=(double)(tmppt->z+z);
    }
    if(dir==1){
      wall_xyz.x=(double)(-tmppt->y+x);
      wall_xyz.y=(double)(tmppt->x+y);
      wall_xyz.z=(double)(tmppt->z+z);
    }
    if(dir==2){
      wall_xyz.x=(double)(-tmppt->x+x);
      wall_xyz.y=(double)(-tmppt->y+y);
      wall_xyz.z=(double)(tmppt->z+z);
    }
    if(dir==3){
      wall_xyz.x=(double)(tmppt->y+x);
      wall_xyz.y=(double)(-tmppt->x+y);
      wall_xyz.z=(double)(tmppt->z+z);
    }
    pos3d(&vis_pos,eye,vec_sub(&tmp,&wall_xyz,pos));
    aPoints[i].x = (int)(vis_pos.x*SCREEN_WIDTH/2)+SCREEN_WIDTH/2;
    aPoints[i].y = (int)(-vis_pos.y*SCREEN_HEIGHT/2)+SCREEN_HEIGHT/2;
    if(vis_pos.z<0)
      back=1;
    if(vis_pos.z>dirmax)
      dirmax=vis_pos.z;
    if(vis_pos.z<dirmin)
      dirmin=vis_pos.z;
    tmppt++;
    }
    if(back==0){
      if(tmppt->typ==1){
        push_polygon(aPoints,i,tft.color565(tmppt->x,tmppt->y,tmppt->z),dirmax,dirmin);
      }
    }
    tmppt++;
  }
  return;
}

void push_polygon(POINT *pntpt, int num, uint16_t col, double dirmax, double dirmin)
{
  int i;
  if(num==0){
    poly_num=0;
    return;
  }
  for(i=0;i<num;i++){
    polygon_dat[poly_num].pnt[i].x=pntpt[i].x;
    polygon_dat[poly_num].pnt[i].y=pntpt[i].y;
  }
  polygon_dat[poly_num].n=num;
  polygon_dat[poly_num].c=col;
  polygon_dat[poly_num].dmax=dirmax;
  polygon_dat[poly_num].dmin=dirmin;
  poly_num++;
  return;
}

void draw_polygon()
{
  int i,j;
  int num;
  double dirmax,dirmin;

  for(i=0;i<poly_num;i++){
    dirmax=0;
    dirmin=0;
    for(j=0;j<poly_num;j++){
      if(polygon_dat[j].n>0){
        if(polygon_dat[j].dmin>dirmin){
          num=j;
          dirmax=polygon_dat[j].dmax;
          dirmin=polygon_dat[j].dmin;
        }else if(polygon_dat[j].dmin==dirmin && polygon_dat[j].dmax>dirmax){
          num=j;
          dirmax=polygon_dat[j].dmax;
          dirmin=polygon_dat[j].dmin;
        }
      }
    }
    if(dirmax==0)
      return;

    Polygon(polygon_dat[num].pnt, polygon_dat[num].n, polygon_dat[num].c);
    polygon_dat[num].n=0;
  }
  return;
}

void Polygon(POINT *pnt, int n, uint16_t col)
{
  int i;

  if (n<3)
    return;
  for(i=2; i<n; i++)
    tft.fillTriangle(SCREEN_START_X+pnt[0].x, SCREEN_START_Y+pnt[0].y,
                     SCREEN_START_X+pnt[i-1].x, SCREEN_START_Y+pnt[i-1].y,
                     SCREEN_START_X+pnt[i].x, SCREEN_START_Y+pnt[i].y, col);
}

struct f_vec *pos3d(struct f_vec *vis,struct f_vec *eye,struct f_vec *obj)
{
  double angx,angy;
  double lenga,lengb,lengc;
  int back;
  back=0;
  lenga=sqrt(eye->x*eye->x+eye->y*eye->y);
  lengb=sqrt(obj->x*obj->x+obj->y*obj->y);
  if(lengb==0)
    angx=0;
  else{
    angx=(eye->x*obj->x+eye->y*obj->y)/lenga/lengb;
    if(angx>=1.0)
    angx=1.0;
    if(angx<=-1.0)
    angx=-1.0;
    angx=acos(angx);
    if(angx>=(M_PI/2-0.01)){
    back=1;
    vis->x=0;
    vis->y=0;
    vis->z=-1;
    return(vis);
    }
  }
  if((-eye->x*obj->y+eye->y*obj->x)<0)
    angx=-angx;
  lengc=sqrt(eye->x*eye->x+eye->y*eye->y);
  lengc=(obj->x*eye->x+obj->y*eye->y)/lengc;
  if(lengc==0){
    if(obj->z>0)
    angy=1;
    else
    angy=-1;
  }else
    angy=obj->z/fabs(lengc);
  vis->x=tan(angx);
  vis->y=angy;
  vis->z=obj->x*obj->x+obj->y*obj->y+obj->z*obj->z;
  return(vis);
}

struct f_vec *vec_add(struct f_vec *result,struct f_vec *para0,struct f_vec *para1)
{
  result->x=para0->x+para1->x;
  result->y=para0->y+para1->y;
  result->z=para0->z+para1->z;
  return(result);
}

struct f_vec *vec_sub(struct f_vec *result,struct f_vec *para0,struct f_vec *para1)
{
  result->x=para0->x-para1->x;
  result->y=para0->y-para1->y;
  result->z=para0->z-para1->z;
  return(result);
}

int load_wall()
{
  File  datFile;
  uint8_t fbuf;
  int num, num_sign;
  int i;
  int no;
  int d0,d1,d2,d3;
  struct f_wall *tmppt;

  if((datFile=SD.open("/wall.dat"))==NULL){
    Serial.print(F("File not found"));
    return(0);
  }
  tmppt=wall_dat;

  num=0;
  num_sign=0;
  i=0;
  while(datFile.available()){
    fbuf=datFile.read();
    if ('0'<=fbuf && fbuf<='9'){
      if (num_sign==0)
        num_sign=1;
      num=num*10+fbuf-'0';
    }else if (fbuf=='-'){
      num_sign=-1;
    }else if (num_sign!=0){
      num=num*num_sign;
      if (i==0)
        d0=num;
      else if (i==1)
        d1=num;
      else if (i==2)
        d2=num;
      else if (i==3)
        d3=num;

      if (i==3){
//Serial.print(d0);
//Serial.print(" ");
//Serial.print(d1);
//Serial.print(" ");
//Serial.print(d2);
//Serial.print(" ");
//Serial.println(d3);
        if(d3==99){
          no=d0;
          wallpt[no]=tmppt;
        }else{
          tmppt->x=d0;
          tmppt->y=d1;
          tmppt->z=d2;
          tmppt->typ=d3;
          tmppt++;
        }
      }

      i++;
      if (i>=4)
        i=0;
      num=0;
      num_sign=0;
    }
  }
  
  datFile.close();
  return(1);
}

int load_maze(int stage_no)
{
  File  datFile;
  char fname[256];
  uint8_t fbuf;
  int num, num_sign;
  int i, j;

  for(i=0;i<22;i++)
    for(j=0;j<16;j++)
      wall[i][j]=10;

  sprintf(fname, "/maze%d.dat", stage_no);
  
  if((datFile=SD.open(fname))==NULL){
    Serial.print(F("File not found"));
    return(0);
  }
  num=0;
  num_sign=0;
  i=0;
  while(datFile.available()){
    fbuf=datFile.read();
    if ('0'<=fbuf && fbuf<='9'){
      if (num_sign==0)
        num_sign=1;
      num=num*10+fbuf-'0';
    }else if (fbuf=='-'){
      num_sign=-1;
    }else if (num_sign!=0){
      num=num*num_sign;
      if (i==0){
        start_x=num;
      }else if (i==1){
        start_y=num+3;
      }else if (i==2){
        goal_x=num;
      }else if (i==3){
        goal_y=num+3;
        my_x=start_x;
        my_y=start_y;
        my_dir=0;
      }else{
        wall[(i-4)/16+3][(i-4) % 16]=num;
      }
      i++;
      num=0;
      num_sign=0;
    }
  }
  
  datFile.close();
  return(1);
}
/*
void tone (uint8_t pin, uint16_t freq, uint16_t leng)
{
  if (tone_init==0){
    ledcSetup(1, 12000, 8);
    ledcAttachPin(pin, 1);
    tone_init=1;
  }
  ledcWriteTone(1, freq);
}

void noTone (uint8_t pin)
{
  ledcWriteTone(1, 1);
}
*/
